package com.ibm.commerce.wc.update.utils;

/* 
 *------------------------------------------------------------------- 
 * IBM Confidential 
 * OCO Source Materials 
 * 
 * WebSphere Commerce 
 * 
 * (c) Copyright International Business Machines Corporation. 
 *     2006, 2007 
 *     All rights reserved. 
 * 
 * The source code for this program is not published or otherwise 
 * divested of its trade secrets, irrespective of what has been 
 * deposited with the US Copyright Office. 
 *------------------------------------------------------------------- 
 */

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.transform.TransformerException;

import org.apache.xerces.parsers.DOMParser;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Attr;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * This class contains static utility functions that manipulates a DOM based on
 * XPath. It also provides general XML utility functions to obtain the defined
 * encoding of an XML file and escape entity references so that they do not get
 * resolved by the XML parser. In addition, it contains helper functions for
 * reading and writing XML files.
 */
public class XMLUpdateUtil {

	/**
	 * IBM copyright notice field.
	 */
	public static final String COPYRIGHT = com.ibm.commerce.copyright.IBMCopyright.SHORT_COPYRIGHT;

	/**
	 * The SAX parser feature for loading external DTD.
	 */
	private static final String XML_FEATURE_LOAD_EXT_DTD = "http://apache.org/xml/features/nonvalidating/load-external-dtd";

	/**
	 * The SAX parser feature for including external general entities.
	 */
	private static final String SAX_FEATURE_EXT_GENERAL_ENT = "http://xml.org/sax/features/external-general-entities";

	/**
	 * The SAX parser feature for including external parameter entities and the
	 * external DTD subset.
	 */
	private static final String SAX_FEATURE_EXT_PARAM_ENT = "http://xml.org/sax/features/external-parameter-entities";

	/**
	 * Regular expression to match attributes in an XML file.
	 */
	private static final String REGEX_ATTR = "([:_A-Za-z][:A-Za-z0-9\\-\\._]*)\\s*=\\s*(\\\"[^\\\"]*\\\"|'[^']*')";

	/**
	 * The <code>encoding</code> attribute name of an XML declaration.
	 */
	public static final String ATTR_NAME_ENC = "encoding";

	/**
	 * The '&' character.
	 */
	private static final String AMP = "&";

	/**
	 * The escaped '&' character.
	 */
	private static final String ESCAPED_AMP = "___AMP___";

	/**
	 * Regular expression to match entity references suffix (after &).
	 */
	private static final String REGEX_ENTITY_REF_SUFFIX = "([:_A-Za-z][:A-Za-z0-9\\-\\._]*);";

	/**
	 * Lookbehind regular expression to not match regular escaped XML characters
	 * as entity references.
	 */
	private static final String REGEX_ESCAPED_XML_CHARS_ENTITY_REF = "(?<!"
			+ AMP + "amp;)(?<!" + AMP + "apos;)(?<!" + AMP + "gt;)(?<!" + AMP
			+ "lt;)(?<!" + AMP + "quot;)";

	/**
	 * Lookbehind regular expression to not match regular escaped XML characters
	 * as unescaped entity references.
	 */
	private static final String REGEX_ESCAPED_XML_CHARS_ESCAPED_ENTITY_REF = "(?<!"
			+ ESCAPED_AMP
			+ "amp;)(?<!"
			+ ESCAPED_AMP
			+ "apos;)(?<!"
			+ ESCAPED_AMP
			+ "gt;)(?<!"
			+ ESCAPED_AMP
			+ "lt;)(?<!"
			+ ESCAPED_AMP
			+ "quot;)";

	/**
	 * Regular expression to match entity references.
	 */
	private static final String REGEX_ENTITY_REF = AMP
			+ REGEX_ENTITY_REF_SUFFIX + REGEX_ESCAPED_XML_CHARS_ENTITY_REF;

	/**
	 * Regular expression to match escaped entity references.
	 */
	private static final String REGEX_ESCAPED_ENTITY_REF = ESCAPED_AMP
			+ REGEX_ENTITY_REF_SUFFIX
			+ REGEX_ESCAPED_XML_CHARS_ESCAPED_ENTITY_REF;

	/**
	 * Regular expression to match XML declaration at the beginning of an XML
	 * file (if any).
	 */
	private static final String REGEX_XML_DECL = "<\\?xml.+?\\?>";

	/**
	 * Byte buffer size for reading file content.
	 */
	private static final int BYTE_BUFFER_SIZE = 8192;

	/**
	 * Appends a child node to the node(s) in a node list.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) to which child node is added.
	 * @param newNode
	 *            The new node to be added.
	 */
	public static void appendChild(Document doc, NodeList matchedNodeList,
			Node newNode) {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			doc.importNode(newNode, true);
			node.appendChild(newNode);
		}
	}

	/**
	 * Appends a child node to the node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) to which the child node is added.
	 * @param newNode
	 *            The new node to be added.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void appendChild(Document doc, String path, Node newNode)
			throws TransformerException, NoXPathMatchException {
		appendChild(doc, selectNodeList(doc, path), newNode);
	}

	/**
	 * Appends a set of nodes to the node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) to which the child node is added.
	 * @param newNodes
	 *            The list of new nodes.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void appendChild(Document doc, String path, NodeList newNodes)
			throws TransformerException, NoXPathMatchException {
		DocumentFragment frag = doc.createDocumentFragment();
		for (int i = 0; i < newNodes.getLength(); i++) {
			frag.appendChild(doc.importNode(newNodes.item(i), true));
		}
		appendChild(doc, path, frag);
	}

	/**
	 * Deletes the list of nodes in a node list.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) to be deleted.
	 */
	public static void delete(Document doc, NodeList matchedNodeList) {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			if (node.getNodeType() == Node.ATTRIBUTE_NODE) {
				((Attr) node).getOwnerElement()
						.removeAttributeNode((Attr) node);
			} else {
				node.getParentNode().removeChild(node);
			}
		}
	}

	/**
	 * Deletes the node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) to be deleted.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void delete(Document doc, String path)
			throws TransformerException, NoXPathMatchException {
		delete(doc, selectNodeList(doc, path));
	}

	/**
	 * Inserts a node before a list of node(s).
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) before which the new node is inserted.
	 * @param newNode
	 *            The new node to be inserted.
	 */
	public static void insertBefore(Document doc, NodeList matchedNodeList,
			Node newNode) {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			doc.importNode(newNode, true);
			node.getParentNode().insertBefore(newNode.cloneNode(true), node);
		}
	}

	/**
	 * Inserts a node before the node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) before which the new node is
	 *            inserted.
	 * @param newNode
	 *            The new node to be inserted.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertBefore(Document doc, String path, Node newNode)
			throws TransformerException, NoXPathMatchException {
		insertBefore(doc, selectNodeList(doc, path), newNode);
	}

	/**
	 * Inserts a set of node before a list of node(s).
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) before which the new nodes are inserted.
	 * @param newNodes
	 *            The list of new nodes.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertBefore(Document doc, NodeList matchedNodeList,
			NodeList newNodes) throws TransformerException,
			NoXPathMatchException {
		DocumentFragment frag = doc.createDocumentFragment();
		for (int i = 0; i < newNodes.getLength(); i++) {
			frag.appendChild(doc.importNode(newNodes.item(i), true));
		}
		insertBefore(doc, matchedNodeList, frag);
	}

	/**
	 * Inserts a set of node before the node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) before which the new nodes are
	 *            inserted.
	 * @param newNodes
	 *            The list of new nodes.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertBefore(Document doc, String path, NodeList newNodes)
			throws TransformerException, NoXPathMatchException {
		DocumentFragment frag = doc.createDocumentFragment();
		for (int i = 0; i < newNodes.getLength(); i++) {
			frag.appendChild(doc.importNode(newNodes.item(i), true));
		}
		insertBefore(doc, path, frag);
	}

	/**
	 * Inserts a node after a list of node(s).
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) after which the new node is inserted.
	 * @param newNode
	 *            The new node to be inserted.
	 * @throws TransformerException
	 */
	public static void insertAfter(Document doc, NodeList matchedNodeList,
			Node newNode) throws TransformerException {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			Node nextSibling = node.getNextSibling();
			doc.importNode(newNode, true);

			if (nextSibling == null) {
				// If the target node has no next sibling, treat the action as
				// an appendChild action on the parent node
				node.getParentNode().appendChild(newNode);
			} else {
				// Otherwise treat it as an insertBefore action on the next
				// sibling
				node.getParentNode().insertBefore(newNode.cloneNode(true),
						nextSibling);
			}
		}
	}

	/**
	 * Inserts a node after the node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node after before which the new node is
	 *            inserted.
	 * @param newNode
	 *            The new node to be inserted.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertAfter(Document doc, String path, Node newNode)
			throws TransformerException, NoXPathMatchException {
		insertAfter(doc, selectNodeList(doc, path), newNode);
	}

	/**
	 * Inserts a set of node after a list of node(s).
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) after which the new nodes are inserted.
	 * @param newNodes
	 *            The list of new nodes.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertAfter(Document doc, NodeList matchedNodeList,
			NodeList newNodes) throws TransformerException,
			NoXPathMatchException {
		DocumentFragment frag = doc.createDocumentFragment();
		for (int i = 0; i < newNodes.getLength(); i++) {
			frag.appendChild(doc.importNode(newNodes.item(i), true));
		}
		insertAfter(doc, matchedNodeList, frag);
	}

	/**
	 * Inserts a set of node after the node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) after which the new nodes are
	 *            inserted.
	 * @param newNodes
	 *            The list of new nodes.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertAfter(Document doc, String path, NodeList newNodes)
			throws TransformerException, NoXPathMatchException {
		DocumentFragment frag = doc.createDocumentFragment();
		for (int i = 0; i < newNodes.getLength(); i++) {
			frag.appendChild(doc.importNode(newNodes.item(i), true));
		}
		insertAfter(doc, path, frag);
	}

	/**
	 * Replace the node(s) in a node list with the new node.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) to be replaced.
	 * @param newNode
	 *            The replacement node.
	 */
	public static void replace(Document doc, NodeList matchedNodeList,
			Node newNode) {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			doc.importNode(newNode, true);
			node.getParentNode().replaceChild(newNode.cloneNode(true), node);
		}
	}

	/**
	 * Replaces the node(s) pointed to by the XPath with the new node.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) to be replaced.
	 * @param newNode
	 *            The replacement node.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void replace(Document doc, String path, Node newNode)
			throws TransformerException, NoXPathMatchException {
		replace(doc, selectNodeList(doc, path), newNode);
	}

	/**
	 * Replaces the node(s) in a list with the new list of nodes.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) to be replaced.
	 * @param newNodes
	 *            The list of replacement nodes.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void replace(Document doc, NodeList matchedNodeList,
			NodeList newNodes) throws TransformerException,
			NoXPathMatchException {
		DocumentFragment frag = doc.createDocumentFragment();
		for (int i = 0; i < newNodes.getLength(); i++) {
			frag.appendChild(doc.importNode(newNodes.item(i), true));
		}
		replace(doc, matchedNodeList, frag);
	}

	/**
	 * Replaces the node(s) pointed to by the XPath with the new list of nodes.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) to be replaced.
	 * @param newNodes
	 *            The list of replacement nodes.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void replace(Document doc, String path, NodeList newNodes)
			throws TransformerException, NoXPathMatchException {
		DocumentFragment frag = doc.createDocumentFragment();
		for (int i = 0; i < newNodes.getLength(); i++) {
			frag.appendChild(doc.importNode(newNodes.item(i), true));
		}
		replace(doc, path, frag);
	}

	/**
	 * Appends a string to the node(s) in a node list. It only works with nodes
	 * of type Attribute, CDATA, Comment, and Text.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) to which value is appended.
	 * @param value
	 *            The value to be appended.
	 * @throws TransformerException
	 */
	public static void appendString(Document doc, NodeList matchedNodeList,
			String value) throws TransformerException {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			if (node.getNodeType() != Node.ATTRIBUTE_NODE
					&& node.getNodeType() != Node.CDATA_SECTION_NODE
					&& node.getNodeType() != Node.COMMENT_NODE
					&& node.getNodeType() != Node.TEXT_NODE) {
				throw new TransformerException(
						"Matched node is not one of the following types: "
								+ "Attribute, CDATA, Comment, Text.");
			} else {
				node.setNodeValue(node.getNodeValue() + value);
			}
		}

	}

	/**
	 * Appends a string to the node(s) pointed to by the XPath. It only works
	 * with nodes of type Attribute, CDATA, Comment, and Text.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param pathThe
	 *            The XPath to the node(s) to which value is appended.
	 * @param value
	 *            The value to be appended.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void appendString(Document doc, String path, String value)
			throws TransformerException, NoXPathMatchException {
		appendString(doc, selectNodeList(doc, path), value);
	}

	/**
	 * Deletes all occurances of a substring from the node(s) pointed to by the
	 * XPath. It only works with nodes of type Attribute, CDATA, Comment, and
	 * Text.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) from which substring is deleted.
	 * @param value
	 *            The substring value to be deleted.
	 * @throws TransformerException
	 */
	public static void deleteSubstring(Document doc, NodeList matchedNodeList,
			String value) throws TransformerException {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			if (node.getNodeType() != Node.ATTRIBUTE_NODE
					&& node.getNodeType() != Node.CDATA_SECTION_NODE
					&& node.getNodeType() != Node.COMMENT_NODE
					&& node.getNodeType() != Node.TEXT_NODE) {
				throw new TransformerException(
						"Matched node is not one of the following types: "
								+ "Attribute, CDATA, Comment, Text.");
			} else {
				if (value.length() != 0) {
					String nodeValue = node.getNodeValue();
					int index = -1;
					while ((index = nodeValue.indexOf(value)) != -1) {
						nodeValue = nodeValue.substring(0, index)
								+ nodeValue.substring(index + value.length());
					}
					node.setNodeValue(nodeValue);
				}
			}
		}

	}

	/**
	 * Deletes all occurances of a substring from the node(s) pointed to by the
	 * XPath. It only works with nodes of type Attribute, CDATA, Comment, and
	 * Text.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the node(s) from which substring is deleted.
	 * @param value
	 *            The substring value to be deleted.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void deleteSubstring(Document doc, String path, String value)
			throws TransformerException, NoXPathMatchException {
		deleteSubstring(doc, selectNodeList(doc, path), value);
	}

	/**
	 * Sets an attribute in the node(s) in a node list. If the attribute does
	 * not exist, a new attribute is added; Otherwise it is set to the new
	 * value.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of element node(s) to which the new attribute is
	 *            added.
	 * @param name
	 *            The new attribute name.
	 * @param value
	 *            The new attribute value.
	 * @throws TransformerException
	 */
	public static void setAttribute(Document doc, NodeList matchedNodeList,
			String name, String value) throws TransformerException {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			if (node.getNodeType() != Node.ELEMENT_NODE) {
				throw new TransformerException(
						"Matched node is not an element.");
			} else {
				((Element) node).setAttribute(name, value);
			}
		}
	}

	/**
	 * Sets an attribute in the node(s) pointed to by the XPath. If the
	 * attribute does not exist, a new attribute is added; Otherwise it is set
	 * to the new value.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param path
	 *            The XPath to the element node(s) to which the new attribute is
	 *            added.
	 * @param name
	 *            The new attribute name.
	 * @param value
	 *            The new attribute value.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void setAttribute(Document doc, String path, String name,
			String value) throws TransformerException, NoXPathMatchException {
		setAttribute(doc, selectNodeList(doc, path), name, value);
	}

	/**
	 * Inserts a comment before the node(s) pointed to by the XPath. If a
	 * matched node is of type Attribute, CDATA, or Text, it will insert the
	 * comment before its parent node.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) before which the comment is added.
	 * @param comment
	 *            The comment text.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertCommentBefore(Document doc,
			NodeList matchedNodeList, String comment)
			throws TransformerException, NoXPathMatchException {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			if (node.getNodeType() == Node.ATTRIBUTE_NODE
					|| node.getNodeType() == Node.CDATA_SECTION_NODE
					|| node.getNodeType() == Node.TEXT_NODE) {
				node = node.getParentNode();
			}
			node.getParentNode().insertBefore(
					doc.createComment(comment).cloneNode(true), node);
		}
	}

	/**
	 * Inserts a comment after the node(s) pointed to by the XPath. If a matched
	 * node is of type Attribute, CDATA, or Text, it will insert the comment
	 * after its parent node.
	 * 
	 * @param doc
	 *            The DOM tree to be modified.
	 * @param matchedNodeList
	 *            The list of node(s) after which the comment is added.
	 * @param comment
	 *            The comment text.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static void insertCommentAfter(Document doc,
			NodeList matchedNodeList, String comment)
			throws TransformerException, NoXPathMatchException {
		for (int i = 0; i < matchedNodeList.getLength(); i++) {
			Node node = matchedNodeList.item(i);
			if (node.getNodeType() == Node.ATTRIBUTE_NODE
					|| node.getNodeType() == Node.CDATA_SECTION_NODE
					|| node.getNodeType() == Node.TEXT_NODE) {
				node = node.getParentNode();
			}
			if (node.getNextSibling() == null) {
				node.getParentNode().appendChild(
						doc.createComment(comment).cloneNode(true));
			} else {
				node.getParentNode().insertBefore(
						doc.createComment(comment).cloneNode(true),
						node.getNextSibling());
			}
		}
	}

	/**
	 * Returns the number of XPath matches in the DOM tree.
	 * 
	 * @param doc
	 *            The DOM tree to be tested against.
	 * @param path
	 *            The XPath to be searched.
	 * @return The number of matches that satisfy <code>path</code> in
	 *         <code>doc</code>.
	 * @throws TransformerException
	 */
	public static int countMatches(Document doc, String path)
			throws TransformerException {
		return XPathAPI.selectNodeList(doc.getDocumentElement(), path)
				.getLength();
	}

	/**
	 * Checks whether a given XPath matches any node in the DOM tree.
	 * 
	 * @param doc
	 *            The DOM tree to be tested against.
	 * @param path
	 *            The XPath to be searched.
	 * @return <code>true</code> if XPath matches; <code>false</code>
	 *         otherwise.
	 * @throws TransformerException
	 */
	public static boolean matches(Document doc, String path)
			throws TransformerException {
		return (countMatches(doc, path) != 0);
	}

	/**
	 * Returns a list of matched node(s) pointed to by the XPath.
	 * 
	 * @param doc
	 *            The document to select node(s) from.
	 * @param path
	 *            The XPath.
	 * @return A list of matched node(s) pointed to by the XPath.
	 * @throws TransformerException
	 * @throws NoXPathMatchException
	 */
	public static NodeList selectNodeList(Document doc, String path)
			throws TransformerException, NoXPathMatchException {
		NodeList nodeList = XPathAPI.selectNodeList(doc.getDocumentElement(),
				path);
		if (nodeList.getLength() == 0) {
			throw new NoXPathMatchException(path);
		}
		return nodeList;
	}

	/**
	 * Returns the encoding of an XML file.
	 * 
	 * @param xmlFile
	 *            The XML file to be parsed.
	 * @return The encoding if declared; <code>null</code> otherwise.
	 * @throws IOException
	 */
	public static String getEncoding(File xmlFile) throws IOException {
		String xmlDecl = getXMLDeclaration(xmlFile);
		if (xmlDecl == null) {
			return null;
		}

		// Use regex to find all attributes in the XML declaration
		Matcher matcher = Pattern.compile(REGEX_ATTR).matcher(xmlDecl);
		while (matcher.find()) {
			// Group 1 is attribute name
			if (matcher.group(1).toLowerCase().equals(ATTR_NAME_ENC)) {
				// Group 2 is attribute value with quotes
				String attr = matcher.group(2);
				return attr.substring(1, attr.length() - 1);
			}
		}
		return null;
	}

	/**
	 * Returns the XML declaration string in an XML file.
	 * 
	 * @param xmlFile
	 *            The XML file to be parsed.
	 * @return The XML declaration string if it exists; <code>null</code>
	 *         otherwise.
	 * @throws IOException
	 */
	public static String getXMLDeclaration(File xmlFile) throws IOException {
		FileInputStream fis = null;
		try {
			int byteRead = -1;
			String line = null;

			// Due to a limitation of the Linux Java SDK, standard file I/O may
			// fail with a MalformedInputException when it hits a non 7-bit
			// ASCII character in a file that has a different encoding than the
			// OS default. This is worked around by reading the file by byte. If
			// the first line contains the XML declaration, is it guaranteed to
			// be 7-bit ASCII. Otherwise it doesn't matter if the encoding for
			// high-byte characters are messed up since the data is not relevant
			// to this method. For more information, refer to:
			// http://www.ibm.com/developerworks/java/jdk/linux/142/runtimeguide.lnx.en.html
			fis = new FileInputStream(xmlFile);
			StringBuffer firstLine = new StringBuffer();
			while ((byteRead = fis.read()) != -1) {
				if (byteRead == (int) '\n' || byteRead == (int) '\r') {
					break;
				} else {
					firstLine.append((char) byteRead);
				}
			}
			Matcher matcher = Pattern.compile(REGEX_XML_DECL,
					Pattern.CASE_INSENSITIVE).matcher(firstLine);
			if (matcher.find()) {
				return matcher.group(0);
			}
			return null;
		} catch (IOException e) {
			throw e;
		} finally {
			if (fis != null) {
				fis.close();
			}
		}
	}

	/**
	 * Escapes entity references in an XML file.
	 * 
	 * @param xmlFile
	 *            The XML file in which entity references are escaped.
	 * @return <code>true</code> if file is modified; <code>false</code>
	 *         otherwise.
	 * @throws IOException
	 */
	public static boolean escapeEntityReferences(File xmlFile)
			throws IOException {
		return escapeEntityReferences(xmlFile, xmlFile);
	}

	/**
	 * Reads content of an input XML file, escapes entity references, and writes
	 * the modified content to another XML file.
	 * 
	 * @param inputFile
	 *            The input XML file to get content from.
	 * @param outputFile
	 *            The output XML file to write modified content to.
	 * @return <code>true</code> if file is modified; <code>false</code>
	 *         otherwise.
	 * @throws IOException
	 */
	public static boolean escapeEntityReferences(File inputFile, File outputFile)
			throws IOException {
		return replaceValues(inputFile, outputFile, REGEX_ENTITY_REF, AMP,
				ESCAPED_AMP);
	}

	/**
	 * Unescapes entity references in an XML file.
	 * 
	 * @param inputFile
	 *            The XML file in which entity references are unescaped.
	 * @return <code>true</code> if file is modified; <code>false</code>
	 *         otherwise.
	 * @throws IOException
	 */
	public static boolean unescapeEntityReferences(File inputFile)
			throws IOException {
		return unescapeEntityReferences(inputFile, inputFile);
	}

	/**
	 * Reads content of an input XML file, unescapes entity references, and
	 * writes the modified content to another XML file.
	 * 
	 * @param inputFile
	 *            The input XML file to get content from.
	 * @param outputFile
	 *            The output XML file to write modified content to.
	 * @return <code>true</code> if file is modified; <code>false</code>
	 *         otherwise.
	 * @throws IOException
	 */
	public static boolean unescapeEntityReferences(File inputFile,
			File outputFile) throws IOException {
		return replaceValues(inputFile, outputFile, REGEX_ESCAPED_ENTITY_REF,
				ESCAPED_AMP, AMP);
	}

	/**
	 * Returns the escaped '&' string.
	 * 
	 * @return The escaped '&' string.
	 */
	public static String getEscapedAmp() {
		return ESCAPED_AMP;
	}

	/**
	 * Reads content of an input XML file, replaces one string with another, and
	 * writes the modified content to another XML file.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @param outputFile
	 *            The output file.
	 * @param regex
	 *            The regular expression to match sections to be escaped.
	 * @param fromStr
	 *            The string to replace.
	 * @param toStr
	 *            The new string for replacement.
	 * @return <code>true</code> if file is modified; <code>false</code>
	 *         otherwise.
	 * @throws IOException
	 */
	private static boolean replaceValues(File inputFile, File outputFile,
			String regex, String fromStr, String toStr) throws IOException {
		String encoding = XMLUpdateUtil.getEncoding(inputFile);
		StringBuffer buffer = new StringBuffer();
		BufferedReader br = null;
		try {
			if (encoding == null) {
				br = new BufferedReader(new FileReader(inputFile));
			} else {
				br = new BufferedReader(new InputStreamReader(
						new FileInputStream(inputFile), encoding));
			}
			char[] byteBuffer = new char[BYTE_BUFFER_SIZE];
			int readSize = 0;
			while ((readSize = br.read(byteBuffer)) != -1) {
				buffer.append(new String(byteBuffer, 0, readSize));
			}
		} catch (IOException e) {
			throw e;
		} finally {
			if (br != null) {
				br.close();
			}
		}

		boolean updated = false;
		Pattern pattern = Pattern.compile(regex);
		Matcher matcher = pattern.matcher(buffer);
		while (matcher.find()) {
			updated = true;
			int startIndex = matcher.start();
			buffer.replace(startIndex, startIndex + fromStr.length(), toStr);

			// Redo regex match after modification to buffer
			matcher = pattern.matcher(buffer);
		}

		if (updated) {
			PrintWriter pw = null;
			try {
				if (encoding == null) {
					pw = new PrintWriter(new OutputStreamWriter(
							new FileOutputStream(outputFile)));
				} else {
					pw = new PrintWriter(new OutputStreamWriter(
							new FileOutputStream(outputFile), encoding));
				}
				pw.write(buffer.toString());
			} catch (IOException e) {
				throw e;
			} finally {
				if (pw != null) {
					pw.close();
				}
			}
			return true;
		}
		return false;
	}

	/**
	 * Reads an XML file and returns its DOM representation. Encoding is
	 * automatically detected, and is set to <code>UTF-8</code> if not defined
	 * in the XML file.
	 * 
	 * @param inputFile
	 *            The input XML file.
	 * @return The DOM representation of the XML file.
	 * @throws IOException
	 * @throws SAXException
	 */
	public static Document readXMLFile(File inputFile) throws IOException,
			SAXException {
		return readXMLFile(inputFile, getEncoding(inputFile));
	}

	/**
	 * Reads an XML file and returns its DOM representation.
	 * 
	 * @param inputFile
	 *            The input file.
	 * @param encoding
	 *            The encoding to be used for reading the XML file. If set to
	 *            <code>null</code>,<code>UTF-8</code> is assumed.
	 * @return The DOM representation of the XML file.
	 * @throws IOException
	 * @throws SAXException
	 */
	public static Document readXMLFile(File inputFile, String encoding)
			throws IOException, SAXException {
		InputSource source = new InputSource(new FileInputStream(inputFile));
		if (encoding != null) {
			source.setEncoding(encoding);
		}

		// Xerces parsers support features to supress exteral DTD and entity
		// loading. For more information, see:
		// http://xerces.apache.org/xerces-j/features.html
		DOMParser dp = new DOMParser();
		dp.setFeature(XML_FEATURE_LOAD_EXT_DTD, false);
		dp.setFeature(SAX_FEATURE_EXT_GENERAL_ENT, false);
		dp.setFeature(SAX_FEATURE_EXT_PARAM_ENT, false);
		dp.parse(source);

		return dp.getDocument();
	}

	/**
	 * Writes a DOM to a file.
	 * 
	 * @param doc
	 *            The DOM.
	 * @param outputFile
	 *            The output XML file.
	 * @param encoding
	 *            The encoding of the output XML file. If set to
	 *            <code>null</code>,<code>UTF-8</code> is assumed.
	 * @param indenting
	 *            Whether to indent the output XML file.
	 * @throws IOException
	 */
	public static void writeXMLFile(Document doc, File outputFile,
			String encoding, boolean indenting) throws IOException {
		OutputFormat of = new OutputFormat();
		of.setIndenting(indenting);
		if (encoding != null) {
			of.setEncoding(encoding);
		}
		writeXMLFile(doc, outputFile, of);
	}

	/**
	 * Writes a DOM to a file.
	 * 
	 * @param doc
	 *            The DOM.
	 * @param outputFile
	 *            The output XML file.
	 * @param of
	 *            The output format.
	 * @throws IOException
	 */
	public static void writeXMLFile(Document doc, File outputFile,
			OutputFormat of) throws IOException {
		if (outputFile.getParentFile() != null
				&& !outputFile.getParentFile().exists()) {
			outputFile.getParentFile().mkdirs();
		}
		FileOutputStream fos = null;
		try {
			fos = new FileOutputStream(outputFile);
			XMLSerializer xs = new XMLSerializer(fos, of);
			xs.asDOMSerializer().serialize(doc);
		} catch (IOException e) {
			throw e;
		} finally {
			if (fos != null) {
				fos.close();
			}
		}
	}

	/**
	 * Serializes a DOM to a string.
	 * 
	 * @param doc
	 *            The DOM.
	 * @param encoding
	 *            The encoding of the output XML file. If set to
	 *            <code>null</code>,<code>UTF-8</code> is assumed.
	 * @param indenting
	 *            Whether to indent the output XML file.
	 * @return The string serialization of the DOM.
	 * @throws IOException
	 */
	public static String serializeToString(Document doc, String encoding,
			boolean indenting) throws IOException {
		OutputFormat of = new OutputFormat();
		of.setIndenting(indenting);
		if (encoding != null) {
			of.setEncoding(encoding);
		}
		return serializeToString(doc, of);
	}

	/**
	 * Serializes a DOM to a string.
	 * 
	 * @param doc
	 *            The DOM.
	 * @param of
	 *            The output format.
	 * @return The string serialization of the DOM.
	 * @throws IOException
	 */
	public static String serializeToString(Document doc, OutputFormat of)
			throws IOException {
		StringWriter sw = new StringWriter();
		XMLSerializer xs = new XMLSerializer(sw, of);
		xs.asDOMSerializer().serialize(doc);
		return sw.getBuffer().toString();
	}

}
