package com.ibm.commerce.wc.update.utils;

/*
 *-----------------------------------------------------------------
 * IBM Confidential
 *
 * OCO Source Materials
 *
 * WebSphere Commerce
 *
 * (C) Copyright IBM Corp. 2006, 2010
 *
 * The source code for this program is not published or otherwise
 * divested of its trade secrets, irrespective of what has
 * been deposited with the U.S. Copyright Office.
 *-----------------------------------------------------------------
 */

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.xerces.parsers.DOMParser;
import org.apache.xpath.XPathAPI;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * This class takes an input configuration file and performs XML update actions
 * accordingly.
 */
public class XMLUpdater {

	/**
	 * IBM copyright notice field.
	 */
	public static final String COPYRIGHT = com.ibm.commerce.copyright.IBMCopyright.SHORT_COPYRIGHT;

	/**
	 * The root <code>xml-update</code> element name.
	 */
	private static final String ELEMENT_NAME_XML_UPDATE = "xml-update";

	/**
	 * The <code>file</code> element name.
	 */
	private static final String ELEMENT_NAME_FILE = "file";

	/**
	 * The <code>input</code> attribute name of the <code>file</code> element.
	 */
	private static final String ATTR_NAME_INPUT = "input";

	/**
	 * The <code>output</code> attribute name of the <code>file</code> element.
	 */
	private static final String ATTR_NAME_OUTPUT = "output";

	/**
	 * The <code>indent</code> attribute name of the <code>file</code> element.
	 */
	private static final String ATTR_NAME_INDENT = "indent";

	/**
	 * The <code>action</code> element name.
	 */
	private static final String ELEMENT_NAME_ACTION = "action";

	/**
	 * The <code>type</code> attribute name of the <code>action</code> element .
	 */
	private static final String ATTR_NAME_TYPE = "type";

	/**
	 * The <code>path</code> attribute name of the <code>action</code> element.
	 */
	private static final String ATTR_NAME_PATH = "path";

	/**
	 * The <code>failOnError</code> attribute name of the <code>action</code>
	 * element.
	 */
	private static final String ATTR_NAME_FAIL_IF_NO_MATCH = "failIfNoMatch";

	/**
	 * The <code>comment</code> attribute name of the <code>action</code>
	 * element.
	 */
	private static final String ATTR_NAME_COMMENT = "comment";

	/**
	 * The <code>skipIfMatchPath</code> attribute name of the
	 * <code>action</code> element.
	 */
	private static final String ATTR_NAME_SKIP_IF_MATCH_PATH = "skipIfMatchPath";

	/**
	 * The <code>name</code> attribute name of the <code>setAttribute</code>
	 * action.
	 */
	private static final String ATTR_NAME_NAME = "name";

	/**
	 * The <code>value</code> attribute name of the <code>setAttribute</code>
	 * action.
	 */
	private static final String ATTR_NAME_VALUE = "value";

	/**
	 * The <code>appendChild</code> action.
	 */
	private static final String ACTION_TYPE_APPEND_CHILD = "appendChild";

	/**
	 * The <code>delete</code> action.
	 */
	private static final String ACTION_TYPE_DELETE = "delete";

	/**
	 * The <code>insertBefore</code> action.
	 */
	private static final String ACTION_TYPE_INSERT_BEFORE = "insertBefore";

	/**
	 * The <code>insertAfter</code> action.
	 */
	private static final String ACTION_TYPE_INSERT_AFTER = "insertAfter";

	/**
	 * The <code>replace</code> action.
	 */
	private static final String ACTION_TYPE_REPLACE = "replace";

	/**
	 * The <code>appendString</code> action.
	 */
	private static final String ACTION_TYPE_APPEND_STR = "appendString";

	/**
	 * The <code>deleteSubstring</code> action.
	 */
	private static final String ACTION_TYPE_DELETE_SUBSTRING = "deleteSubstring";

	/**
	 * The <code>setAttribute</code> action.
	 */
	private static final String ACTION_TYPE_SET_ATTR = "setAttribute";

	/**
	 * The prefix of comment for tagging XML updates by this tool.
	 */
	private static final String XML_UPDATE_COMMENT_PREFIX = "[XMLUpdate:"
			+ (new SimpleDateFormat("yyMMddhhmmss").format(new Date(System
					.currentTimeMillis()))) + "]";

	/**
	 * The DOM tree representing the XML config file.
	 */
	private Document fCfgDoc = null;

	/**
	 * The directory in which the input and output files are located, if they
	 * are given as relative paths.
	 */
	private File fDir = null;

	/**
	 * The path to the input file. When set in single file mode, it will
	 * override the value defined in the configuration XML.
	 */
	private String fInputPath = null;

	/**
	 * The path to the output file. When set in single file mode, it will
	 * override the value defined in the configuration XML.
	 */
	private String fOutputPath = null;

	/**
	 * Toggle on whether the tool should support only one &lt;file&gt; element.
	 */
	private boolean fLimitToSingleFile = false;

	/**
	 * The Ant task for logging purposes as used by the Ant wrapper.
	 */
	private Task fTask = null;

	/**
	 * Constructor.
	 * 
	 * @param cfgDoc
	 *            The DOM tree representing the XML config file.
	 */
	public XMLUpdater(Document cfgDoc) {
		fCfgDoc = cfgDoc;
	}

	/**
	 * Constructor.
	 * 
	 * @param cfgDoc
	 *            The DOM tree representing the XML config file.
	 * @param dir
	 *            The directory in which the input and output files are located.
	 */
	public XMLUpdater(Document cfgDoc, File dir) {
		this(cfgDoc);
		fDir = dir;
	}

	/**
	 * Constructor.
	 * 
	 * @param cfgFile
	 *            The XML config file.
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public XMLUpdater(File cfgFile) throws IOException,
			ParserConfigurationException, SAXException {
		DOMParser dp = new DOMParser();
		dp.parse(new InputSource(new FileInputStream(cfgFile)));
		fCfgDoc = dp.getDocument();
	}

	/**
	 * Constructor.
	 * 
	 * @param cfgFile
	 *            The XML config file.
	 * @param dir
	 *            The directory in which the input and output files are located.
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 */
	public XMLUpdater(File cfgFile, File dir) throws IOException,
			ParserConfigurationException, SAXException {
		this(cfgFile);
		fDir = dir;
	}

	/**
	 * Sets the path to the input file. When set in single file mode, it will
	 * override the value defined in the configuration XML.
	 * 
	 * @param inputPath
	 *            The path to the input file.
	 */
	public void setInputPath(String inputPath) {
		fInputPath = inputPath;
	}

	/**
	 * Sets the path to the output file. When set in single file mode, it will
	 * override the value defined in the configuration XML.
	 * 
	 * @param outputPath
	 *            The path to the output file.
	 */
	public void setOutputPath(String outputPath) {
		fOutputPath = outputPath;
	}

	/**
	 * Sets whether the tool should be limited to single file.
	 * 
	 * @param limitToSingleFile
	 */
	public void setLimitToSingleFile(boolean limitToSingleFile) {
		fLimitToSingleFile = limitToSingleFile;
	}

	/**
	 * Sets the Ant task for logging purposes.
	 * 
	 * @param task
	 *            The Ant task.
	 */
	public void setTask(Task task) {
		fTask = task;
	}

	/**
	 * Performs all updates as per input file.
	 * 
	 * @throws TransformerException
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws NoSuchActionException
	 * @throws NoXPathMatchException
	 */
	public void updateAll() throws TransformerException, IOException,
			ParserConfigurationException, SAXException, NoSuchActionException,
			NoXPathMatchException {
		NodeList fileList = XPathAPI.selectNodeList(fCfgDoc
				.getDocumentElement(), "/" + ELEMENT_NAME_XML_UPDATE + "/"
				+ ELEMENT_NAME_FILE);
		if (fLimitToSingleFile && fileList.getLength() > 1) {
			throw new SAXException("Only one <file> element is allowed.");
		}
		for (int i = 0; i < fileList.getLength(); i++) {
			updateFile((Element) fileList.item(i));
		}
	}

	/**
	 * Determines and updates a file.
	 * 
	 * @param fileElement
	 *            The XML element specifying the file to be modified and actions
	 *            to be taken.
	 * @throws IOException
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws TransformerException
	 * @throws NoSuchActionException
	 * @throws NoXPathMatchException
	 */
	private void updateFile(Element fileElement) throws IOException,
			ParserConfigurationException, SAXException, TransformerException,
			NoSuchActionException, NoXPathMatchException {
		boolean indent = false;
		String inputPath = fileElement.getAttribute(ATTR_NAME_INPUT);

		// Override input path if it is set
		if (fLimitToSingleFile && fInputPath != null) {
			inputPath = fInputPath;
		}
		String outputPath = inputPath;
		if (fileElement.hasAttribute(ATTR_NAME_OUTPUT)) {
			outputPath = fileElement.getAttribute(ATTR_NAME_OUTPUT);
		}

		// Override output path if it is set
		if (fLimitToSingleFile && fOutputPath != null) {
			outputPath = fOutputPath;
		}
		if (fileElement.hasAttribute(ATTR_NAME_INDENT)
				&& (new Boolean(fileElement.getAttribute(ATTR_NAME_INDENT)))
						.booleanValue()) {
			indent = true;
		}

		File inputFile = null;
		if (fDir != null && !new File(inputPath).isAbsolute()) {
			inputFile = new File(fDir, inputPath);
		} else {
			inputFile = new File(inputPath);
		}

		boolean escapedEntityRefs = XMLUpdateUtil
				.escapeEntityReferences(inputFile);
		if (escapedEntityRefs) {
			logInfo("Entity references detected and escaped from "
					+ inputFile.getAbsolutePath() + ".");
		}

		logInfo("Updating " + inputFile.getAbsolutePath() + ".");
		String encoding = XMLUpdateUtil.getEncoding(inputFile);
		Document doc = XMLUpdateUtil.readXMLFile(inputFile, encoding);

		NodeList fileElementChildList = fileElement.getChildNodes();
		for (int i = 0; i < fileElementChildList.getLength(); i++) {
			if (fileElementChildList.item(i).getNodeType() == Node.ELEMENT_NODE
					&& fileElementChildList.item(i).getNodeName().equals(
							ELEMENT_NAME_ACTION)) {
				performAction((Element) fileElementChildList.item(i), doc);
			}
		}

		// Only if all actions can be performed successfully would the DOM tree
		// be available for serialization
		File outputFile = null;
		if (fDir != null && !new File(outputPath).isAbsolute()) {
			outputFile = new File(fDir, outputPath);
		} else {
			outputFile = new File(outputPath);
		}
		if (outputFile.getParentFile() != null
				&& !outputFile.getParentFile().exists()) {
			outputFile.getParentFile().mkdirs();
		}
		logInfo("Serializing output to " + outputFile.getAbsolutePath() + ".");
		XMLUpdateUtil.writeXMLFile(doc, outputFile, encoding, indent);

		if (escapedEntityRefs) {
			logInfo("Unescaping entity references in "
					+ outputFile.getAbsolutePath() + ".");
			XMLUpdateUtil.unescapeEntityReferences(outputFile);
		}
	}

	/**
	 * Determines and performs an action.
	 * 
	 * @param actionElement
	 *            The XML element specifying an action.
	 * @param doc
	 *            The DOM tree to be modified.
	 * @throws TransformerException
	 * @throws NoSuchActionException
	 * @throws NoXPathMatchException
	 */
	private void performAction(Element actionElement, Document doc)
			throws TransformerException, NoSuchActionException,
			NoXPathMatchException {
		boolean failIfNoMatch = false;
		String type = actionElement.getAttribute(ATTR_NAME_TYPE);
		String path = actionElement.getAttribute(ATTR_NAME_PATH);
		if (actionElement.hasAttribute(ATTR_NAME_FAIL_IF_NO_MATCH)
				&& (new Boolean(actionElement
						.getAttribute(ATTR_NAME_FAIL_IF_NO_MATCH)))
						.booleanValue()) {
			failIfNoMatch = true;
		}
		String skipIfMatchPath = null;
		if (actionElement.hasAttribute(ATTR_NAME_SKIP_IF_MATCH_PATH)) {
			skipIfMatchPath = actionElement
					.getAttribute(ATTR_NAME_SKIP_IF_MATCH_PATH);
		}
		if (skipIfMatchPath != null && skipIfMatchPath.length() != 0) {
			if (XMLUpdateUtil.matches(doc, skipIfMatchPath)) {
				logWarning("Conditional path matches, action will be skipped: "
						+ skipIfMatchPath);
				return;
			}
		}
		String comment = null;
		if (actionElement.hasAttribute(ATTR_NAME_COMMENT)) {
			comment = actionElement.getAttribute(ATTR_NAME_COMMENT);
		}

		try {
			if (type.equals(ACTION_TYPE_APPEND_CHILD)) {
				logInfo("Performing " + ACTION_TYPE_APPEND_CHILD + " action.");
				if (comment != null) {
					XMLUpdateUtil
							.appendChild(doc, path, doc
									.createComment(XML_UPDATE_COMMENT_PREFIX
											+ comment));
				} else {
					XMLUpdateUtil.appendChild(doc, path, actionElement
							.getChildNodes());
				}
			} else if (type.equals(ACTION_TYPE_DELETE)) {
				logInfo("Performing " + ACTION_TYPE_DELETE + " action.");
				if (comment != null) {
					NodeList matchedNodeList = XMLUpdateUtil.selectNodeList(
							doc, path);
					XMLUpdateUtil.insertCommentBefore(doc, matchedNodeList,
							XML_UPDATE_COMMENT_PREFIX + comment);
					XMLUpdateUtil.delete(doc, matchedNodeList);
				} else {
					XMLUpdateUtil.delete(doc, path);
				}
			} else if (type.equals(ACTION_TYPE_INSERT_BEFORE)) {
				logInfo("Performing " + ACTION_TYPE_INSERT_BEFORE + " action.");
				if (comment != null) {
					NodeList matchedNodeList = XMLUpdateUtil.selectNodeList(
							doc, path);
					XMLUpdateUtil.insertCommentBefore(doc, matchedNodeList,
							XML_UPDATE_COMMENT_PREFIX + comment);
					XMLUpdateUtil.insertBefore(doc, matchedNodeList,
							actionElement.getChildNodes());
				} else {
					XMLUpdateUtil.insertBefore(doc, path, actionElement
							.getChildNodes());
				}
			} else if (type.equals(ACTION_TYPE_INSERT_AFTER)) {
				logInfo("Performing " + ACTION_TYPE_INSERT_AFTER + " action.");
				if (comment != null) {
					NodeList matchedNodeList = XMLUpdateUtil.selectNodeList(
							doc, path);
					XMLUpdateUtil.insertAfter(doc, matchedNodeList,
							actionElement.getChildNodes());
					XMLUpdateUtil.insertCommentAfter(doc, matchedNodeList,
							XML_UPDATE_COMMENT_PREFIX + comment);
				} else {
					XMLUpdateUtil.insertAfter(doc, path, actionElement
							.getChildNodes());
				}
			} else if (type.equals(ACTION_TYPE_REPLACE)) {
				logInfo("Performing " + ACTION_TYPE_REPLACE + " action.");
				if (comment != null) {
					NodeList matchedNodeList = XMLUpdateUtil.selectNodeList(
							doc, path);
					XMLUpdateUtil.insertCommentBefore(doc, matchedNodeList,
							XML_UPDATE_COMMENT_PREFIX + comment);
					XMLUpdateUtil.replace(doc, matchedNodeList, actionElement
							.getChildNodes());
				} else {
					XMLUpdateUtil.replace(doc, path, actionElement
							.getChildNodes());
				}
			} else if (type.equals(ACTION_TYPE_APPEND_STR)) {
				logInfo("Performing " + ACTION_TYPE_APPEND_STR + " action.");
				String value = actionElement.getAttribute(ATTR_NAME_VALUE);
				if (comment != null) {
					NodeList matchedNodeList = XMLUpdateUtil.selectNodeList(
							doc, path);
					XMLUpdateUtil.insertCommentBefore(doc, matchedNodeList,
							XML_UPDATE_COMMENT_PREFIX + comment);
					XMLUpdateUtil.appendString(doc, matchedNodeList, value);
				} else {
					XMLUpdateUtil.appendString(doc, path, value);
				}
			} else if (type.equals(ACTION_TYPE_DELETE_SUBSTRING)) {
				logInfo("Performing " + ACTION_TYPE_DELETE_SUBSTRING
						+ " action.");
				String value = actionElement.getAttribute(ATTR_NAME_VALUE);
				if (comment != null) {
					NodeList matchedNodeList = XMLUpdateUtil.selectNodeList(
							doc, path);
					XMLUpdateUtil.insertCommentBefore(doc, matchedNodeList,
							XML_UPDATE_COMMENT_PREFIX + comment);
					XMLUpdateUtil.deleteSubstring(doc, matchedNodeList, value);
				} else {
					XMLUpdateUtil.deleteSubstring(doc, path, value);
				}
			} else if (type.equals(ACTION_TYPE_SET_ATTR)) {
				logInfo("Performing " + ACTION_TYPE_SET_ATTR + " action.");
				String name = actionElement.getAttribute(ATTR_NAME_NAME);
				String value = actionElement.getAttribute(ATTR_NAME_VALUE);
				if (comment != null) {
					NodeList matchedNodeList = XMLUpdateUtil.selectNodeList(
							doc, path);
					XMLUpdateUtil.insertCommentBefore(doc, matchedNodeList,
							XML_UPDATE_COMMENT_PREFIX + comment);
					XMLUpdateUtil.setAttribute(doc, matchedNodeList, name,
							value);
				} else {
					XMLUpdateUtil.setAttribute(doc, path, name, value);
				}
			} else {
				throw new NoSuchActionException(type);
			}
		} catch (NoXPathMatchException e) {
			if (failIfNoMatch) {
				throw e;
			} else {
				logWarning("Path does not match, action will be skipped: "
						+ e.getMessage());
			}
		}
	}

	/**
	 * Helper method to log a regular message.
	 * 
	 * @param message
	 *            The message to be logged.
	 */
	private void logInfo(String message) {
		if (fTask != null) {
			fTask.log(message);
		} else {
			System.out.println(message);
		}
	}

	/**
	 * Helper method to log a warning message.
	 * 
	 * @param message
	 *            The messag to be logged.
	 */
	private void logWarning(String message) {
		if (fTask != null) {
			fTask.log("WARNING: " + message, Project.MSG_WARN);
		} else {
			System.err.println(message);
		}
	}

}